import {
  empty, assign, extend, size, entries, keys, values, map, reduce
} from '../../src/lang';

import assert from 'assert';

describe('lang', () => {

  describe('.empty()', () => {

    const TestClass = function() {};
    TestClass.prototype = {constructor: TestClass, foo: 1};

    it('false is empty', () => expect(empty(false)).to.equal(true));
    it('true is not empty', () => expect(empty(true)).to.equal(false));

    it('null is empty', () => expect(empty(null)).to.equal(true));
    it('undefined is empty', () => expect(empty(undefined)).to.equal(true));

    it('zero-length string is empty', () => expect(empty('')).to.equal(true));
    it('non-zero length string is not empty', () => expect(empty(' ')).to.equal(false));

    it('zero-length array is empty', () => expect(empty([])).to.equal(true));
    it('non-zero length array is not empty', () => expect(empty([0])).to.equal(false));

    it('integer zero is empty', () => expect(empty(0)).to.equal(true));
    it('number other than integer zero is not empty', () => expect(empty(-1)).to.equal(false));
    it('NaN is not empty', () => expect(empty(NaN)).to.equal(false));

    it('Map with no keys is empty', () => expect(empty(new Map())).to.equal(true));
    it('Set with no values is empty', () => expect(empty(new Set())).to.equal(true));

    it('object that owns no properties is empty', () => {
      assert(Object.getOwnPropertyNames({}).length === 0);
      expect(empty({})).to.equal(true);
    });

    it('object that owns at least one property is not empty', () => {
      let TestClass = function() {};

      TestClass.prototype = {constructor: TestClass};

      let obj = new TestClass();
      obj.bar = 2;
      expect(empty(obj)).to.equal(false);
    });

    it('Map with at least one key is not empty', () => {
      let map = new Map([['foo', 1]]);
      expect(empty(map)).to.equal(false);
    });

    it('Set with at least one value is not empty', () => {
      let set = new Set(['foo']);
      expect(empty(set)).to.equal(false);
    });
  });


  describe('.assign()', () => {
    it('assigns donor properties to a recipient', () => {
      let recipient = {foo: 1, bar: 'two'}
      let donor1 = {bar: 2, baz: 3};
      let donor2 = {bar: 'II'};
      let output = assign(recipient, donor1, donor2);

      expect(output).to.equal(recipient);
      expect(output.foo).to.equal(recipient.foo);
      expect(output.bar).to.equal(donor2.bar);
      expect(output.baz).to.equal(donor1.baz);
      expect(size(output)).to.equal(3);
    });
  });

  describe('.entries()', () => {
    it('converts an Array', () => {
      let {array: input} = fixture('iterables-1');

      // right-pad the array with undefined
      input.length = 4;

      let output = entries(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(4);

      expect(output[0]).to.be.an.instanceof(Array);
      expect(output[1]).to.be.an.instanceof(Array);

      expect(output[0].length).to.equal(2);
      expect(output[1].length).to.equal(2);
      expect(output[0][0]).to.equal(0);
      expect(output[0][1]).to.equal(4);
      expect(output[1][0]).to.equal(1);
      expect(output[1][1]).to.equal(5);
      expect(output[2][0]).to.equal(2);
      expect(output[2][1]).to.equal(undefined);
      expect(output[3][0]).to.equal(3);
      expect(output[3][1]).to.equal(undefined);
    });

    it('converts an Object', () => {
      let {object: input} = fixture('iterables-1');
      let output = entries(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(2);

      expect(output[0]).to.be.an.instanceof(Array);
      expect(output[1]).to.be.an.instanceof(Array);

      expect(output[0].length).to.equal(2);
      expect(output[1].length).to.equal(2);
      expect(output[0][0]).to.equal('foo');
      expect(output[0][1]).to.equal('one');
      expect(output[1][0]).to.equal('bar');
      expect(output[1][1]).to.equal(2);
    });

    it('converts a Map', () => {
      let {map: input} = fixture('iterables-1');
      let output = entries(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(2);

      expect(output[0]).to.be.an.instanceof(Array);
      expect(output[1]).to.be.an.instanceof(Array);

      expect(output[0].length).to.equal(2);
      expect(output[1].length).to.equal(2);
      expect(output[0][0]).to.equal('foo');
      expect(output[0][1]).to.equal('one');
      expect(output[1][0]).to.equal('bar');
      expect(output[1][1]).to.equal(2);
    });

    it('converts a Set', () => {
      let {set: input} = fixture('iterables-1');
      let output = entries(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(2);

      expect(output[0]).to.be.an.instanceof(Array);
      expect(output[1]).to.be.an.instanceof(Array);

      expect(output[0].length).to.equal(2);
      expect(output[1].length).to.equal(2);
      expect(output[0][0]).to.equal('foo');
      expect(output[0][1]).to.equal('foo');
      expect(output[1][0]).to.equal('bar');
      expect(output[1][1]).to.equal('bar');
    });

  });

  describe('.keys()', () => {
    it('works on an Array', () => {
      let {array: input} = fixture('iterables-1');
      let output = keys(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(4);
      expect(output[0]).to.equal(0);
      expect(output[1]).to.equal(1);
      expect(output[2]).to.equal(2);
      expect(output[3]).to.equal(3);
    });

    it('works on an Object', () => {
      let {object: input} = fixture('iterables-1');
      let output = keys(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(2);
      expect(output[0]).to.equal('foo');
      expect(output[1]).to.equal('bar');
    });

    it('works on a Map', () => {
      let {map: input} = fixture('iterables-1');
      let output = keys(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(2);
      expect(output[0]).to.equal('foo');
      expect(output[1]).to.equal('bar');
    });

    it('works on a Set', () => {
      let {set: input} = fixture('iterables-1');
      let output = keys(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(2);
      expect(output[0]).to.equal('foo');
      expect(output[1]).to.equal('bar');
    });
  });

  describe('.values()', () => {
    it('works on an Array', () => {
      let {array: input} = fixture('iterables-1');
      let output = values(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(4);
      expect(output[0]).to.equal(4);
      expect(output[1]).to.equal(5);
      expect(output[2]).to.equal(undefined);
      expect(output[3]).to.equal(undefined);
    });

    it('works on an Object', () => {
      let {object: input} = fixture('iterables-1');
      let output = values(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(2);
      expect(output[0]).to.equal('one');
      expect(output[1]).to.equal(2);
    });

    it('works on a Map', () => {
      let {map: input} = fixture('iterables-1');
      let output = values(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(2);
      expect(output[0]).to.equal('one');
      expect(output[1]).to.equal(2);
    });

    it('works on a Set', () => {
      let {set: input} = fixture('iterables-1');
      let output = values(input);

      expect(output).to.be.an.instanceof(Array);
      expect(output.length).to.equal(2);
      expect(output[0]).to.equal('foo');
      expect(output[1]).to.equal('bar');
    });

  });

  describe('.size()', () => {
    it('works on an Array', () => {
      let {array: input} = fixture('iterables-1');
      expect(size(input)).to.equal(4);
    });

    it('works on an Object', () => {
      let {object: input} = fixture('iterables-1');
      expect(size(input)).to.equal(2);
    });

    it('works on a Map', () => {
      let {map: input} = fixture('iterables-1');
      expect(size(input)).to.equal(2);
    });

    it('works on a Set', () => {
      let {set: input} = fixture('iterables-1');
      expect(size(input)).to.equal(2);
    });

  });

  describe('.reduce()', () => {

    it('works on an Array', () => {
      let {array: input} = fixture('sequences-1');

      let sum = (result, v, k, input) => {
        expect(v).to.equal(input[k]);
        return result + v;
      };

      expect(reduce(input, sum, 0)).to.equal(45);
    });

    it('works on an Object', () => {
      let {object: input} = fixture('sequences-1');

      let sum = (result, v, k, input) => {
        expect(v).to.equal(input[k]);
        return result + v;
      };

      expect(reduce(input, sum, 0)).to.equal(45);
    })

    it('works on a Map', () => {
      let {map: input} = fixture('sequences-1');

      let sum = (result, v, k, input) => {
        expect(v).to.equal(input.get(k));
        return result + v;
      };

      expect(reduce(input, sum, 0)).to.equal(45);
    });

    it('works on a Set', () => {
      let {set: input} = fixture('sequences-1');

      let sum = (result, v, k, input) => {
        expect(v).to.equal(k);
        expect(input.has(k)).to.be.ok;
        return result + v;
      };

      expect(reduce(input, sum, 0)).to.equal(45);
    });
  });


  describe('.reduce.until()', () => {

    it('works on an Array', () => {
      let {array: input} = fixture('sequences-1');

      let sum = (result, v, k, input) => {
        expect(v).to.equal(input[k]);
        return result + v;
      };

      let test = (result, v, k, input) => v >= 9;

      expect(reduce.until(input, sum, 0, test)).to.equal(36);
    });

    it('works on an Object', () => {
      let {object: input} = fixture('sequences-1');

      let sum = (result, v, k, input) => {
        expect(v).to.equal(input[k]);
        return result + v;
      };

      let test = (result, v, k, input) => v >= 9;

      expect(reduce.until(input, sum, 0, test)).to.equal(36);
    })

    it('works on a Map', () => {
      let {map: input} = fixture('sequences-1');

      let sum = (result, v, k, input) => {
        expect(v).to.equal(input.get(k));
        return result + v;
      };

      let test = (result, v, k, input) => v >= 9;

      expect(reduce.until(input, sum, 0, test)).to.equal(36);
    });

    it('works on a Set', () => {
      let {set: input} = fixture('sequences-1');

      let sum = (result, v, k, input) => {
        expect(v).to.equal(k);
        expect(input.has(k)).to.be.ok;
        return result + v;
      };

      let test = (result, v, k, input) => v >= 9;

      expect(reduce.until(input, sum, 0, test)).to.equal(36);
    });

  });


  describe('.map()', () => {

    it('works on an Array', () => {
      let {array: input} = fixture('sequences-1');

      let squared = (v, k, input) => {
        expect(v).to.equal(input[k]);
        return v * v;
      };

      let expected = [0, 1, 4, 9, 16, 25, 36, 49, 64, 81];

      expect(map(input, squared)).to.include(...expected);
    });

    it('works on an Object', () => {
      let {object: input} = fixture('sequences-1');

      let squared = (v, k, input) => {
        expect(v).to.equal(input[k]);
        return v * v;
      };

      let expected = [0, 1, 4, 9, 16, 25, 36, 49, 64, 81];

      expect(map(input, squared)).to.include(...expected);
    })

    it('works on a Map', () => {
      let {map: input} = fixture('sequences-1');

      let squared = (v, k, input) => {
        expect(v).to.equal(input.get(k));
        return v * v;
      };

      let expected = [0, 1, 4, 9, 16, 25, 36, 49, 64, 81];

      expect(map(input, squared)).to.include(...expected);
    });

    it('works on a Set', () => {
      let {set: input} = fixture('sequences-1');

      let squared = (v, k, input) => {
        expect(input.has(k)).to.be.ok;
        expect(v).to.equal(k);
        return v * v;
      };

      let expected = [0, 1, 4, 9, 16, 25, 36, 49, 64, 81];

      expect(map(input, squared)).to.include(...expected);
    });

  });

});
